import torch
from tqdm import tqdm
from torch.optim import Adam, SGD
from sklearn.metrics import f1_score,roc_auc_score
from model_new import SubNet,LMF,TextSubNet,Net_LSTM,Net_FC
from HPFN import FNL2NO
from data_new import load_mosi,load_iemocap
import numpy as np
import torch.nn as nn
import os
from util import get_index_3, mig_loss_function_multi, hgr_loss_function
import torch.nn.functional as F
import argparse
from torch.utils.data import DataLoader, Dataset


torch.manual_seed(123)
torch.cuda.manual_seed_all(123)
#TODO Hypers: bs,alpha,lr
def main(args):

    data_path = args.path
    train_set, valid_set, test_set, label_set, input_dims = load_iemocap(data_path, 'neutral', args)
    batch_sz = 32
    train_loader_label = DataLoader(label_set, shuffle=True, batch_size=batch_sz)
    train_loader_all = DataLoader(train_set, shuffle=True, batch_size=bs)
    dev_loader = DataLoader(valid_set, shuffle=False, batch_size=batch_sz * 3)
    test_loader = DataLoader(test_set, shuffle=False, batch_size=batch_sz * 3)
    weight_decay = 0.

    model_dict = dict()
    optimizer_dict = dict()
    model_test_dict = dict()
    if args.lmf:
        r = 8
        ahid = 32
        vhid = 16
        thid = 100
        thid_2 = thid // 2
        input_dims = (74, 35, 300)
        adr = 0.1
        tdr = 0.1
        vdr = 0.1
        output_dim = 2
        model_lmf = LMF(input_dims, (ahid, vhid, thid), thid_2, (adr, vdr, tdr, 0.5), output_dim, r).cuda()
        optimizer_lmf = Adam(model_lmf.parameters(), lr=1e-3, weight_decay=weight_decay)

    if args.hgr:
        r = 8
        ahid = 32
        vhid = 32
        thid = 100
        thid_2 = 32
        input_dims = (74, 35, 300)
        adr = 0.1
        tdr = 0.1
        vdr = 0.1
        output_dim = 2
        model_hgr = LMF(input_dims, (ahid, vhid, thid), thid_2, (adr, vdr, tdr, 0.5), output_dim, r, hgr=True).cuda()
        optimizer_hgr = Adam(model_hgr.parameters(), lr=1e-3, weight_decay=weight_decay)

    model_dict['text'] = Net_LSTM(input_dims[2], hidden_size=100).cuda()
    model_dict['visual'] = Net_FC(input_dims[1],visual=True).cuda()
    model_dict['audio'] = Net_FC(input_dims[0]).cuda()

    for k in model_dict.keys():
        if k == 'visual':
            optimizer_dict[k] = Adam(model_dict[k].parameters(), lr=1e-3, weight_decay=weight_decay)
        else:
            optimizer_dict[k] = Adam(model_dict[k].parameters(), lr=1e-3, weight_decay=weight_decay)


    best_valid_loss = {'text': 1000, 'visual': 1000, 'audio': 1000}
    best_test_auc = {'text': 0.0, 'visual': 0.0, 'audio': 0.0}
    best_valid = 1000
    p = cal_prior(train_loader_all)
    print("Prior:", p)
    for e in range(args.epoch):
        if args.hgr:
            train_hgr(e, model_hgr, optimizer_hgr, train_loader_label, args.alpha)
            valid_auc, valid_loss = test_hpfn(model_hgr, dev_loader,hgr=True)
            test_auc, test_loss = test_hpfn(model_hgr, test_loader,hgr=True)
            if valid_loss < best_valid:
                best_valid = valid_loss
                best_test = test_auc
            print("Test auc:",best_test)
            continue
        if args.lmf:
            train_hpfn(e, model_lmf, optimizer_lmf, train_loader_label)
            valid_auc, valid_loss = test_hpfn(model_lmf, dev_loader)
            test_auc, test_loss = test_hpfn(model_lmf, test_loader)
            if valid_loss < best_valid:
                best_valid = valid_loss
                best_test = test_auc
            print("Test auc:", best_test)
            continue
        if args.s:
            train(e, model_dict, optimizer_dict, train_loader_label, p,alpha=0, mig=False)
        if args.tcgm:
            train(e, model_dict, optimizer_dict, train_loader_label, p,alpha=0, mig=False)

            train(e, model_dict, optimizer_dict, train_loader_all, p,alpha=args.alpha, mig=True)
        current_valid_auc, current_valid_loss = test(model_dict, dev_loader,p)
        current_test_auc, test_loss = test(model_dict, test_loader,p)
        for k in best_test_auc.keys():
            if current_valid_loss[k] < best_valid_loss[k]:
                best_valid_loss[k] = current_valid_loss[k]
                best_test_auc[k] = current_test_auc[k]
                if args.tcgm:
                    torch.save(model_dict[k], 'model_' + k + '.std')
                if args.s:
                    torch.save(model_dict[k], 'models_' + k + '.std')
    if args.tcgm:
        model_test_dict['text'] = torch.load('model_text.std').cuda()
        model_test_dict['visual'] = torch.load('model_visual.std').cuda()
        model_test_dict['audio'] = torch.load('model_audio.std').cuda()
        auc, _ = test(model_test_dict, test_loader,p)
        print("MIG All:", auc)
    if args.s:
        model_test_dict['text'] = torch.load('models_text.std').cuda()
        model_test_dict['visual'] = torch.load('models_visual.std').cuda()
        model_test_dict['audio'] = torch.load('models_audio.std').cuda()
        auc, _ = test(model_test_dict, test_loader,p)
        print("S  All:", auc)


def test_hpfn(model, loader, hgr=False):
    criterion = nn.CrossEntropyLoss()
    model.eval()
    y_true = []
    y_pred = {'all': []}
    train_iter = tqdm(loader)
    train_loss = 0.0
    for batch in train_iter:
        a, v, t, y = batch
        batch_size = y.size(0)
        if batch_size == 1:
            continue
        t, v, a, y = t.cuda().float(), v.cuda().float().squeeze(), a.cuda().float().squeeze(), y.cuda()
        y = y.view(-1, 2).long()
        y = torch.max(y,1)[1]
        if hgr:
            y_tilde,_,_,_ = model(a, v, t)
        else:
            y_tilde = model(a, v, t)
        y_pred['all'].append(F.softmax(y_tilde,1).detach().cpu().numpy())
        y_true.append(y.detach().cpu().numpy())
        loss = criterion(y_tilde, y)
        train_loss += loss.item()
    train_loss = train_loss / len(loader.dataset)
    y_true = np.concatenate(y_true, axis=0)
    auc = dict()
    for k in y_pred.keys():
        y_pred[k] = np.concatenate(y_pred[k], axis=0)[:, 1]
        #y_pred[k] = np.argmax(y_pred[k], 1)
        auc[k] = roc_auc_score(y_true,y_pred[k])
    return auc, train_loss

def test(model_dict, loader,p):
    criterion = nn.CrossEntropyLoss()
    for k in model_dict.keys():
        model_dict[k].eval()
    y_true = []
    y_pred = {'text': [], 'visual': [], 'audio': [], 'all':[]}
    loss_collect = {'text': 0.0, 'visual': 0.0, 'audio': 0.0, 'all': 0.0}
    with torch.no_grad():
        for batch in loader:
            a, v, t, y = batch
            t, v, a, y = t.cuda().float(), v.cuda().float().squeeze(), a.cuda().float().squeeze(), y.cuda()
            y_tilde_t = model_dict['text'](t)
            y_tilde_v = model_dict['visual'](v)
            y_tilde_a = model_dict['audio'](a)
            y = y.view(-1, 2).long()
            y = torch.max(y, 1)[1]
            y_pred['text'].append(F.softmax(y_tilde_t, 1).detach().cpu().numpy())
            y_pred['visual'].append(F.softmax(y_tilde_v, 1).detach().cpu().numpy())
            y_pred['audio'].append(F.softmax(y_tilde_a, 1).detach().cpu().numpy())
            y_true.append(y.detach().cpu().numpy())
            output_dict = dict()
            output_dict[1], output_dict[2], output_dict[3] = F.softmax(y_tilde_t, dim=1), F.softmax(y_tilde_v,dim=1), F.softmax(y_tilde_a, dim=1)
            y_pred['all'].append(F.softmax(p * output_dict[1] * output_dict[2] * output_dict[3],1).detach().cpu().numpy())
            loss_collect['text'] += criterion(y_tilde_t, y)
            loss_collect['audio'] += criterion(y_tilde_a, y)
            loss_collect['visual'] += criterion(y_tilde_v, y)
    loss_collect['all'] = loss_collect['text'] + loss_collect['audio'] + loss_collect['visual']
    y_true = np.concatenate(y_true, axis=0)
    auc = dict()
    for k in y_pred.keys():
        y_pred[k] = np.concatenate(y_pred[k], axis=0)[:, 1]
        #y_pred[k] = np.argmax(y_pred[k], 1)
        auc[k] = roc_auc_score(y_true,y_pred[k])
    return auc, loss_collect

def train_hgr(epoch,model,optimizer, loader, alpha):
    criterion = nn.CrossEntropyLoss(size_average=False)
    model.train()
    y_true = []
    y_pred = {'all': []}
    train_iter = tqdm(loader)
    train_loss = 0.0
    for batch in train_iter:

        optimizer.zero_grad()
        a, v, t, y = batch
        batch_size = y.size(0)
        if batch_size == 1:
            continue
        t, v, a, y = t.cuda().float(), v.cuda().float().squeeze(), a.cuda().float().squeeze(), y.cuda()

        y = y.view(-1, 2).long()
        y = torch.max(y, 1)[1]
        y_tilde, fa, fv, ft = model(a, v, t)
        y_pred['all'].append(y_tilde.detach().cpu().numpy())
        y_true.append(y.detach().cpu().numpy())
        loss = criterion(y_tilde, y) + alpha * (hgr_loss_function(fa,fv)+hgr_loss_function(fa,ft)+hgr_loss_function(fv,ft))
        loss.backward()

        optimizer.step()
        train_iter.set_description(f"Epoch {epoch}, current batch loss: {loss.item() / batch_size}")
        train_loss += loss.item()
    y_true = np.concatenate(y_true, axis=0)




def train(epoch,model_dict,optimizer_dict, loader, p, alpha, mig):
    criterion = nn.CrossEntropyLoss(size_average=False)
    for k in model_dict.keys():
        model_dict[k].train()
    y_true = []
    y_pred = {'text': [], 'visual': [], 'audio': [], 'all': []}
    train_iter = tqdm(loader)
    train_loss = 0.0
    d = get_index_3(bs)
    for batch in train_iter:
        for k in optimizer_dict.keys():
            optimizer_dict[k].zero_grad()
        a, v, t, y = batch
        batch_size = y.size(0)
        if batch_size == 1:
            continue
        t, v, a, y = t.cuda().float(), v.cuda().float().squeeze(), a.cuda().float().squeeze(), y.cuda()

        y = y.view(-1, 2).long()
        y = torch.max(y, 1)[1]

        y_tilde_t = model_dict['text'](t)
        y_tilde_v = model_dict['visual'](v)
        y_tilde_a = model_dict['audio'](a)
        y_pred['text'].append(F.softmax(y_tilde_t,1).detach().cpu().numpy())
        y_pred['visual'].append(F.softmax(y_tilde_v,1).detach().cpu().numpy())
        y_pred['audio'].append(F.softmax(y_tilde_a,1).detach().cpu().numpy())
        y_true.append(y.detach().cpu().numpy())
        output_dict = dict()
        output_dict[1], output_dict[2], output_dict[3] = F.softmax(y_tilde_t, dim=1), F.softmax(y_tilde_v, dim=1), F.softmax(y_tilde_a, dim=1)
        y_pred['all'].append((output_dict[1] * output_dict[2] * output_dict[3]).detach().cpu().numpy())
        if batch_size != bs:
            d_ = get_index_3(batch_size)
        else:
            d_ = d
        if mig:
            loss = alpha * mig_loss_function_multi(output_dict, modal_num=3, bs=batch_size, d=d_, p=p)
        else:
            loss = criterion(y_tilde_t, y) + criterion(y_tilde_v, y) + criterion(y_tilde_a, y)
        loss.backward()
        for k in optimizer_dict.keys():
            optimizer_dict[k].step()
        train_iter.set_description(f"Epoch {epoch}, current batch loss: {loss.item() / batch_size}")
        train_loss += loss.item()



def train_hpfn(epoch, model, optimizer, loader):
    criterion = nn.CrossEntropyLoss()
    model.train()
    y_true = []
    y_pred = { 'all': []}
    train_iter = tqdm(loader)
    train_loss = 0.0
    for batch in train_iter:
        optimizer.zero_grad()
        a, v, t, y = batch
        batch_size = y.size(0)
        if batch_size == 1:
            continue
        t, v, a, y = t.cuda().float(), v.cuda().float().squeeze(), a.cuda().float().squeeze(), y.cuda()
        y = y.view(-1, 2).long()
        y = torch.max(y, 1)[1]
        y_tilde = model(a,v,t)
        y_pred['all'].append(y_tilde.detach().cpu().numpy())
        y_true.append(y.detach().cpu().numpy())
        loss = criterion(y_tilde, y)
        loss.backward()
        optimizer.step()
        train_iter.set_description(f"Epoch {epoch}, current batch loss: {round(loss.item() / batch_size, 4)}")
        train_loss += loss.item()
    train_loss = train_loss / len(loader.dataset)
    print(f"Training loss: {round(train_loss, 4)}")


def cal_prior(loader):

    y_true = []
    train_iter = tqdm(loader)
    for batch in train_iter:
        a, v, t, y = batch
        batch_size = y.size(0)
        if batch_size == 1:
            continue
        t, v, a, y = t.cuda().float(), v.cuda().float().squeeze(), a.cuda().float().squeeze(), y.cuda().squeeze()
        y = y.view(-1, 2).long()
        y = torch.max(y, 1)[1]
        y_true.append(y.detach().cpu().numpy())

    y_true = np.concatenate(y_true, axis=0)
    return torch.FloatTensor([(y_true==0).astype(np.int16).sum() / len(y_true), (y_true==1).astype(np.int16).sum() / len(y_true)]).cuda()

if __name__ == '__main__':
    parser = argparse.ArgumentParser(description='MOSI')
    parser.add_argument('--device', type=int)
    parser.add_argument('--epoch', type=int, default=20, help="Number of epochs to train.")
    parser.add_argument('--tcgm', default=False, action='store_true', help='MIG or not')
    parser.add_argument('--hpfn', default=False, action='store_true', help='HPFN or not')
    parser.add_argument('--lmf', default=False, action='store_true', help='LMF or not')
    parser.add_argument('--hgr', default=False, action='store_true', help='HGR or not')
    parser.add_argument('--s', default=False, action='store_true', help='supervised or not')
    parser.add_argument('--alpha', type=float, default=1.0, help='hyper-parameters for MIG')
    parser.add_argument('--r', type=float, default=1.0, help='label fraction')
    parser.add_argument('--seed', type=int, help='random seed')
    parser.add_argument('--bs', type=int, help='TGCM batch size')
    parser.add_argument('--path', type=str, help='The path to data')

    args = parser.parse_args()
    bs = args.bs
    print("Current bs:{}, alpha:{}".format(bs,args.alpha))
    main(args)
