import numpy as np
import sklearn
import matplotlib.pyplot as plt
from sklearn import datasets
from scipy.stats import ortho_group
import scipy
import torch
from main import args

def sigmoid(x):
    return 1. / (1+np.exp(x))
seed = 5678
np.random.seed(seed)
scipy.random.seed(seed)
sample_num = 15000
real_sample_num = 10000
num_epoch = 100
ratio = args.ratio
print("Ratio:",ratio)
view_train = dict()
view_test = dict()
view_val = dict()
random_orth = ortho_group.rvs(2)

M_1_1 = []
M_1_0 = []
X_1_p = np.array([[1,1], [2,1]])
Y_1_p = np.array([1,0])
index = np.random.randint(len(X_1_p), size=sample_num)
X_1 = X_1_p[index]
Y_1 = Y_1_p[index]
X_1 = sigmoid((X_1 * 5).dot(random_orth)) + np.random.normal(size=X_1.shape) * ratio
M_1_1 += X_1[np.where(Y_1==1)].tolist()
M_1_0 += X_1[np.where(Y_1==0)].tolist()
M_1_1 = np.array(M_1_1)[:5000]
M_1_0 = np.array(M_1_0)[:5000]
print(M_1_1.shape, M_1_0.shape)

# random 2 by 2 orthognal matrix
random_orth = ortho_group.rvs(2)
M_2_1 = []
M_2_0 = []
X_2_p = np.array([[1,-1], [0,-1]])
Y_2_p = np.array([1,0])
index = np.random.randint(len(X_2_p), size=sample_num)
X_2 = X_2_p[index]
Y_2 = Y_2_p[index]
X_2 = sigmoid((X_2 * 5).dot(random_orth)) + np.random.normal(size=X_2.shape) * ratio
M_2_1 += X_2[np.where(Y_2==1)].tolist()
M_2_0 += X_2[np.where(Y_2==0)].tolist()
M_2_1 = np.array(M_2_1)[:5000]
M_2_0 = np.array(M_2_0)[:5000]


random_orth = ortho_group.rvs(2)
M_3_1 = []
M_3_0 = []
X_3_p = np.array([[-1,-1], [-2,-1]])
Y_3_p = np.array([1,0])
index = np.random.randint(len(X_3_p), size=sample_num)
X_3 = X_3_p[index]
Y_3 = Y_3_p[index]
X_3 = sigmoid((X_3 * 5).dot(random_orth)) + np.random.normal(size=X_3.shape) * ratio
M_3_1 += X_3[np.where(Y_3==1)].tolist()
M_3_0 += X_3[np.where(Y_3==0)].tolist()
M_3_1 = np.array(M_3_1)[:5000]
M_3_0 = np.array(M_3_0)[:5000]

label = np.zeros((real_sample_num))
label[: real_sample_num // 2] = 1
seed = 12
np.random.seed(seed)
np.random.shuffle(label)

M_1 = np.concatenate((M_1_1,M_2_1,M_3_1),axis=1)
M_0 = np.concatenate((M_1_0,M_2_0,M_3_0),axis=1)
M = np.concatenate((M_1,M_0), axis=0)
print("label shape:", label.shape, "data shape:", M.shape)
np.random.seed(seed)
np.random.shuffle(M)

training_split = int(real_sample_num * 0.5)
val_split = int(real_sample_num * 0.7)
print("Train split:{}, Val split:{}".format(training_split,val_split))
view_train[1] = torch.from_numpy(M[:, :2]).float()[:training_split]
view_test[1] = torch.from_numpy(M[:, :2]).float()[training_split:val_split]
view_val[1] = torch.from_numpy(M[:, :2]).float()[val_split:]
view_train[2] = torch.from_numpy(M[:, 2:4]).float()[:training_split]
view_test[2] = torch.from_numpy(M[:, 2:4]).float()[training_split:val_split]
view_val[2] = torch.from_numpy(M[:, 2:4]).float()[val_split:]
view_train[3] = torch.from_numpy(M[:, 4:]).float()[:training_split]
view_test[3] = torch.from_numpy(M[:, 4:]).float()[training_split:val_split]
view_val[3] = torch.from_numpy(M[:, 4:]).float()[val_split:]
label_train = label[:training_split]
label_test = label[training_split:val_split]
label_val = label[val_split:]

split = int(len(label_train) * args.f)
print("labeled number :", split)

if args.cluster:
    from sklearn.cluster import KMeans, SpectralClustering
    kmeans = KMeans(n_clusters=2, random_state=0).fit(torch.cat((view_train[1], view_train[2], view_train[3]), dim=1))
    pred = kmeans.predict(torch.cat((view_test[1], view_test[2], view_test[3]), dim=1))
    acc = (pred == label_test).astype(np.float32).sum() / len(pred)
    print("Acc (Kmeans++):", max(acc, 1.0 - acc))

    train_ = torch.cat((view_train[1], view_train[2], view_train[3]), dim=1)
    test_ = torch.cat((view_test[1], view_test[2], view_test[3]), dim=1)
    cat = torch.cat((train_, test_), dim=0)
    clustering = SpectralClustering(n_clusters=2, assign_labels="kmeans", random_state=0).fit(cat)
    pred = clustering.labels_[len(view_train[1]):]
    acc = (pred == label_test).astype(np.float32).sum() / len(pred)
    print("ACC (SpectralC):", max(acc, 1.0 - acc))
