import argparse
import sys
import torch.nn.functional as F
import torch
from tqdm import tqdm

sys.path.append('..')
sys.path.append('.')
from utils.vat import VATLoss

parser = argparse.ArgumentParser(description='20Text')
parser.add_argument('--device', type=int)
parser.add_argument('--epoch', type=int,default=100,help="Number of epochs to train.")
parser.add_argument('--modalities', type=int, default=3, help='Number of modalities')
parser.add_argument('--classes', type=int, default=2, help='Number of classes')
parser.add_argument('--split', type=int, default=4000, help='Number of feature split')
parser.add_argument('--labeled', type=int, default=500, help='Number of labeled data')
parser.add_argument('--thresh', type=float, default=0.9, help='Threshold')
parser.add_argument('--label_fraction', type=float, default=0.1, help='label_fraction')
parser.add_argument('--mig', type=int, default=0, help='MIG or not')
parser.add_argument('--s', type=int, default=0, help='MIG or not')
parser.add_argument('--hgr', type=int, default=0, help='HGR or not')
parser.add_argument('--vat', type=int, default=0, help='VAT or not')
parser.add_argument('--alpha', type=float, default=1.0, help='VAT or not')
parser.add_argument('--lam', type=float, default=1.0, help='VAT hyper-parameters')
parser.add_argument('--xi', type=float, default=0.01, help='VAT perturbation( best hyper is default)')
parser.add_argument('--lr', type=float, default=1e-2, help='Learning rate')
parser.add_argument('--data', type=str, help='dataset')
parser.add_argument('--seed', type=int, help='random seed')

args = parser.parse_args()

if args.data == 'Cornell' or args.data == 'Texas' or args.data == 'Washington' or args.data == 'Wisconsin':
    args.classes = 5
    args.modalities = 2


from dataset.dataprepare_Reuters import *
from models.model import *
from utils.utils import mig_loss_function,enu_modalities,mig_loss_function_U,get_index_3,mig_loss_function_multi,mig_loss_function_multi_W,hgr_loss_function,js_loss_function_multi

best_val_acc = np.zeros(args.modalities+1)
best_test_acc = np.zeros(args.modalities+1)
best_test_acc_all = 0
best_val_acc_all = 0

def test(epoch,device):

    acc = 0.0
    acc_ = np.zeros(args.modalities+1)
    for idx, (data_dict,label) in enumerate(dataloader['test']):

        label = label.long().to(device)
        output_total = 0.0
        for i in range(1, 1 + args.modalities):

            output = Net[i](data_dict[i].to(device))
            output_total += output

        pred = torch.max(output_total, 1)[1]
        acc += pred.eq(label.data).cpu().sum()

        for i in range(1, 1 + args.modalities):
            pred = torch.max(Net[i](data_dict[i].to(device)), 1)[1]
            acc_[i] += pred.eq(label.data).cpu().sum()

    acc = float(acc) / float(len(dataloader['test'].dataset))
    #print(acc_)
    #print(float(len(dataloader['test'].dataset)))
    #acc = acc_.sum() / (float(len(dataloader['test'].dataset)) * args.modalities)
    #print("-- Test Epoch:{}, Best Acc = {:.2%}".format(epoch, best_acc))

    return acc_/float(len(dataloader['test'].dataset)), acc

def test_hgr(epoch,device):

    acc = 0.0
    for i in range(1, 1 + args.modalities):
        Net[i].eval()
    for idx, (data_dict,label) in enumerate(dataloader['test']):

        label = label.long().to(device)

        output_collect = []
        for i in range(1, 1 + args.modalities):
            output = Net[i](data_dict[i].to(device), hgr=True)
            output_collect.append(output)

        feature = torch.cat(output_collect, dim=1)
        output = H_Linear(feature)

        pred = torch.max(output, 1)[1]
        acc += pred.eq(label.data).cpu().sum()

    acc = float(acc) / float(len(dataloader['test'].dataset))

    #print("-- Test Epoch:{}, Best Acc = {:.2%}".format(epoch, best_acc))
    return acc
def val(epoch,device):

    acc = 0.0
    acc_ = np.zeros(args.modalities+1)
    for idx, (data_dict,label) in enumerate(dataloader['val']):

        label = label.long().to(device)

        output_total = 0.0
        for i in range(1, 1 + args.modalities):

            output = Net[i](data_dict[i].to(device))
            output_total += output

        pred = torch.max(output_total, 1)[1]
        acc += pred.eq(label.data).cpu().sum()
        for i in range(1, 1 + args.modalities):
            pred = torch.max(Net[i](data_dict[i].to(device)), 1)[1]
            acc_[i] += pred.eq(label.data).cpu().sum()

    acc = float(acc) / float(len(dataloader['val'].dataset))
    #acc = acc_.sum() / (float(len(dataloader['val'].dataset))*args.modalities)
    #print("-- val Epoch:{}, Best Acc = {:.2%}".format(epoch, best_acc))
    return acc_ / float(len(dataloader['val'].dataset)), acc

def VAT(epoch,device,alpha):

    train_loss = 0.0
    vat_loss = VATLoss(xi=0.1, eps=0.1, ip=1)

    for idx, (data_dict, label) in enumerate(dataloader['vat']):

        loss = 0.0
        for i in range(1,1 + args.modalities):
            Optimizer[i].zero_grad()

        for i in range(1, 1 + args.modalities):

            loss += alpha * vat_loss(Net[i], data_dict[i].to(device))

        loss.backward()

        for i in range(1, 1 + args.modalities):
            Optimizer[i].step()

        train_loss += loss.data



    print(">> VAT Train Epoch:{}".format(epoch))

def MIG_train_Sequence(epoch, device,alpha):

    train_loss = 0.0
    p = torch.FloatTensor([0.25,0.25,0.25,0.25]).to(device)

    acc = np.zeros((args.modalities),dtype=np.float32)
    pairs = enu_modalities(args.modalities)
    pairs = [[1,2,3],[1,3,2],[2,1,3],[2,3,1],[3,2,1],[3,1,2]]
    for data_dict, label in dataloader['mig']:
        for idx,(i,j,k) in enumerate(pairs):
            ldata, rdata, kdata = data_dict[i].to(device), data_dict[j].to(device), data_dict[k].to(device)
            label = label.long().to(device)

            Optimizer[i].zero_grad()
            Optimizer[j].zero_grad()
            output_i = torch.exp(Net[i](ldata))
            output_j = torch.exp(Net[j](rdata))

            loss = alpha * mig_loss_function_U(output_i, output_j, p, len(label), device=device)
            loss.backward()

            Optimizer[i].step()
            Optimizer[j].step()

            Optimizer[i].zero_grad()
            Optimizer[j].zero_grad()
            Optimizer[k].zero_grad()
            output_i = torch.exp(Net[i](ldata))
            output_j = torch.exp(Net[j](rdata))
            output_k = torch.exp(Net[k](kdata))

            output = (output_i * output_j) / (output_i * output_j).sum(1).unsqueeze(1)
            loss = alpha * mig_loss_function_U(output, output_k, p, len(label), device=device)
            loss.backward()

            Optimizer[i].step()
            Optimizer[j].step()
            Optimizer[k].step()

def MIG_train_Pair(epoch, device,alpha):

    train_loss = 0.0
    p = [1.0 / float(args.classes) for i in range(args.classes)]
    p = torch.FloatTensor(p).to(device)


    pairs = enu_modalities(args.modalities)
    for data_dict, label in dataloader['mig']:
        for idx,(i,j) in enumerate(pairs):
            ldata, rdata = data_dict[i].to(device), data_dict[j].to(device)
            label = label.long().to(device)

            Optimizer[i].zero_grad()
            Optimizer[j].zero_grad()
            output_i = torch.exp(Net[i](ldata))
            output_j = torch.exp(Net[j](rdata))

            loss = alpha *  mig_loss_function(output_i,output_j,p,len(label),device=device)
            loss.backward()
            Optimizer[i].step()
            Optimizer[j].step()

    #Testing........
    acc = 0.0
    for data_dict, label in dataloader['mig']:

        output_total = 0.0
        label = label.long().to(device)
        for i in range(1, 1 + args.modalities):

            output = Net[i](data_dict[i].to(device))
            output_total += output

        pred = torch.max(output_total, 1)[1]
        acc += pred.eq(label.data).cpu().sum()

    acc = float(acc)/float(len(dataloader['mig'].dataset))
    print(">> MIG Train Epoch:{}, Acc {:.4f}".format(epoch, acc))

def MIG_train_U(epoch, device,alpha):

    train_loss = 0.0
    p = [1.0/float(args.classes) for i in range(args.classes)]
    p = torch.FloatTensor(p).to(device)
    n = int((args.modalities * (args.modalities - 1)) /2)
    acc = np.zeros((n),dtype=np.float32)
    pairs = enu_modalities(args.modalities)
    output_dict = dict()
    d = get_index_3(32)
    for data_dict, label in tqdm(dataloader['mig']):
        if len(label) != 32:
            continue
        for i in range(1, 1 + args.modalities):
            Optimizer[i].zero_grad()
            output = Net[i](data_dict[i].to(device), mig=True)
            output_dict[i] = output

        loss = -alpha * mig_loss_function_multi(output_dict=output_dict, modal_num=args.modalities, p=p, bs=len(label), device=device, sampling_num=1000, d=d)
        loss.backward()
        train_loss += loss.data
        for i in range(1, 1 + args.modalities):
            Optimizer[i].step()

        output_dict.clear()

def HDR_train(epoch,device):

    for t in range(epoch):

        acc = np.zeros((args.modalities), dtype=np.float32)
        train_loss = 0.0
        pairs = enu_modalities(args.modalities)
        for i in range(1, 1 + args.modalities):
            Net[i].train()

        for data_dict, label in dataloader['train']:
            output_collect = []
            label = label.long().to(device)
            for i in range(1, 1 + args.modalities):

                Optimizer[i].zero_grad()
            H_Optimizer.zero_grad()

            for i in range(1, 1 + args.modalities):
                output = Net[i](data_dict[i].to(device),hgr = True)
                output_collect.append(output)

            feature = torch.cat(output_collect,dim=1)
            output = H_Linear(feature)
            loss = (1 - args.lam) * F.nll_loss(output, label)
            loss.backward()
            pred = torch.max(output, 1)[1]
            acc += pred.eq(label.data).cpu().sum()
            train_loss += loss.data

            for i in range(1, 1 + args.modalities):
                Optimizer[i].step()
            H_Optimizer.step()

        print("Training acc:",acc/float(len(dataloader['train'].dataset)))
        for data_dict, label in dataloader['mig']:
            for idx, (i, j) in enumerate(pairs):
                ldata, rdata = data_dict[i].to(device), data_dict[j].to(device)
                label = label.long().to(device)

                Optimizer[i].zero_grad()
                Optimizer[j].zero_grad()
                output_i = Net[i](ldata,hgr=True)
                output_j = Net[j](rdata,hgr=True)
                loss = args.lam * hgr_loss_function(output_i,output_j,device=device)
                loss.backward()
                Optimizer[i].step()
                Optimizer[j].step()

        global best_test_acc
        acc = test_hgr(t, device)
        best_test_acc = max(best_test_acc,acc)
        print("B:",best_test_acc)


def train(epoch, device):

    for t in range(epoch):

        acc = np.zeros((args.modalities), dtype=np.float32)
        train_loss = 0.0

        for idx, (data_dict,label) in enumerate(dataloader['train']):

            label = label.long().to(device)

            for i in range(1, 1 + args.modalities):
                Scheduler[i].step()
                Optimizer[i].zero_grad()
                output = Net[i](data_dict[i].to(device))
                loss = F.nll_loss(output, label)
                loss.backward()
                Optimizer[i].step()

                pred = torch.max(output, 1)[1]
                acc[i-1] += pred.eq(label.data).cpu().sum()
                train_loss += loss.data


        acc /= float(len(dataloader['train'].dataset))
        print("Train Set:",float(len(dataloader['train'].dataset)))
        #train_loss = float(train_loss)/ float(len(train_loader.dataset))
        print("S Train Epoch:{}, Acc 1= {:.2%}, Acc 2 = {:.2%}".format(t, acc[0],acc[1]))

        if args.vat:
            VAT(t,device,args.alpha)


        elif args.mig and acc[0] > args.thresh and acc[1]>args.thresh and acc[2]>args.thresh :

            MIG_train_U(t, device, args.alpha)
        global best_val_acc
        global best_test_acc
        global best_test_acc_all
        global best_val_acc_all
        val_acc, val_total_acc = val(t,device)
        for i in range(1,args.modalities+1):

            if val_acc[i] > best_val_acc[i]:
                best_val_acc[i] = val_acc[i]
                acc_, acc = test(t,device)
                best_test_acc[i] = acc_[i]
                print("i: Best test acc:{}, Best all acc:{}".format(best_test_acc[i], best_test_acc_all))
            if val_total_acc > best_val_acc_all:
                best_val_acc_all = val_total_acc
                acc_, acc = test(t, device)
                best_test_acc_all = max(best_test_acc_all, acc)



    print("Final acc:",best_test_acc)
    print("Final all acc:", best_test_acc_all)



if __name__ == '__main__':

    if args.mig == 1 or args.vat == 1 or args.s == 1:
        train(args.epoch, args.device)
    elif args.hgr == 1:
        HDR_train(args.epoch,args.device)
